@@ -1,6 +1,6 @@
 use v6;
 
-unit module TAP:ver<0.3.4>;
+unit module TAP:ver<0.3.6>;
 
 role Entry {
     has Str:D $.raw is required handles <Str>;
@@ -40,11 +40,9 @@ class Sub-Test is Test {
         my @plans = @!entries.grep(Plan);
         if !@plans {
             @errors.push: "Subtest $usable-number doesn't have a plan";
-        }
-        elsif @plans > 1 {
+        } elsif @plans > 1 {
             @errors.push: "Subtest $usable-number has multiple plans";
-        }
-        elsif @plans[0].tests != @tests.elems {
+        } elsif @plans[0].tests != @tests.elems {
             @errors.push: "Subtest $usable-number expected { @plans[0].tests } but contains { @tests.elems } tests";
         }
         @errors;
@@ -57,6 +55,9 @@ class Bailout does Entry {
 class Comment does Entry {
     has Str:D $.comment is required;
 }
+class Pragma does Entry {
+    has Bool:D %.identifiers is required;
+}
 class YAML does Entry {
     has Str:D $.serialized is required;
     has Any $.deserialized;
@@ -68,16 +69,9 @@ my role Entry::Handler {
     method handle-entry(Entry) { ... }
     method end-entries() { }
     method listen(Supply $supply) {
-        $supply.act(-> $entry {
-                self.handle-entry($entry);
-            },
-            done => {
-                self.end-entries();
-            },
-            quit => {
-                self.end-entries();
-            }
-        );
+        my $act = { self.handle-entry($^entry) };
+        my $done = my $quit = { self.end-entries() };
+        $supply.act($act, :$done, :$quit);
     }
 }
 
@@ -104,11 +98,11 @@ class Result {
         $!exit-status.defined ?? ($!exit-status.exitcode +< 8) +| $!exit-status.signal !! Int;
     }
 
-    method has-problems($ignore-exit) {
-        @!failed || @!errors || (!$ignore-exit && self.exit-failed);
+    method has-problems($ignore-exit = False) {
+        so @!failed || @!errors || (!$ignore-exit && self.exit-failed);
     }
     method exit-failed() {
-        $!exit-status.defined && ($!exit-status.exitcode || $!exit-status.signal);
+        so $!exit-status.defined && ($!exit-status.exitcode || $!exit-status.signal);
     }
 }
 
@@ -157,10 +151,10 @@ class Aggregator {
 
 
     method has-problems() {
-        $!todo-passed || self.has-errors;
+        so $!todo-passed || self.has-errors;
     }
     method has-errors() {
-        $!failed + $!errors + $!exit-failed;
+        so $!failed + $!errors + $!exit-failed;
     }
     method get-status() {
         self.has-errors || $!tests-run != $!passed ?? 'FAILED' !! $!tests-run ?? 'PASS' !! 'NOTESTS';
@@ -169,7 +163,7 @@ class Aggregator {
 
 my grammar Grammar {
     regex TOP {
-        ^ [ <plan> | <test> | <bailout> | <version> | <comment> || <unknown> ] $
+        ^ [ <plan> | <test> | <bailout> | <version> | <pragma> | <comment> || <unknown> ] $
     }
     token sp { <[\s] - [\n]> }
     token num { <[0..9]>+ }
@@ -197,6 +191,12 @@ my grammar Grammar {
     token version {
         :i 'TAP version ' <version=.num>
     }
+    token pragma-identifier {
+        $<sign>=<[+-]> $<name>=[<alnum>+]
+    }
+    token pragma {
+        'pragma ' <pragma-identifier>+ % ' '
+    }
     token comment {
         '#' <.sp>* $<comment>=[\N*]
     }
@@ -206,7 +206,7 @@ my grammar Grammar {
         <.indent($indent)> '  ...'
     }
     token sub-entry(Int $indent) {
-        <plan> | <test> | <comment> | <yaml($indent)> | <sub-test($indent)> || <!before <.sp> > <unknown>
+        <plan> | <test> | <comment> | <pragma> | <yaml($indent)> | <sub-test($indent)> || <!before <.sp> > <unknown>
     }
     token indent(Int $indent) {
         '    ' ** { $indent }
@@ -250,6 +250,13 @@ my grammar Grammar {
         method version($/) {
             make TAP::Version.new(:raw(~$/), :version(+$<version>));
         }
+        method pragma-identifier($/) {
+            make $<name> => ?( $<sign> eq '+' );
+        }
+        method pragma($/) {
+            my Bool:D %identifiers = @<pragma-identifier>.map(*.ast);
+            make TAP::Pragma.new(:raw(~$/), :%identifiers);
+        }
         method comment($/) {
             make TAP::Comment.new(:raw(~$/), :comment(~$<comment>));
         }
@@ -309,29 +316,24 @@ my sub parser(Supply $input --> Supply) {
             if $mode == Normal {
                 if $line ~~ / ^ '  ---' / {
                     set-state(Yaml, $line);
-                }
-                elsif $line ~~ &indented {
+                } elsif $line ~~ &indented {
                     set-state(SubTest, $line);
-                }
-                else {
+                } else {
                     emit-reset $grammar.parse($line);
                 }
             }
             elsif $mode == SubTest {
                 if $line ~~ &indented {
                     @buffer.push: $line;
-                }
-                elsif $grammar.parse($line, :rule('test')) -> $test {
+                } elsif $grammar.parse($line, :rule('test')) -> $test {
                     my $raw = (|@buffer, $line).join("\n");
                     if $grammar.parse($raw, :rule('sub-test')) -> $subtest {
                         emit-reset $subtest;
-                    }
-                    else {
+                    } else {
                         emit-unknown;
                         emit-reset $test;
                     }
-                }
-                else {
+                } else {
                     emit-unknown $line;
                 }
             }
@@ -342,13 +344,11 @@ my sub parser(Supply $input --> Supply) {
                         my $raw = @buffer.join("\n");
                         if $grammar.parse($raw, :rule('yaml')) -> $yaml {
                             emit-reset $yaml;
-                        }
-                        else {
+                        } else {
                             emit-unknown;
                         }
                     }
-                }
-                else {
+                } else {
                     emit-unknown $line;
                 }
             }
@@ -452,8 +452,7 @@ class Formatter::Text does Formatter {
                     if $result.wait -> $wait {
                         if $result.exit {
                             $output ~= self.format-failure("Non-zero exit status: { $result.exit }\n");
-                        }
-                        else {
+                        } else {
                             $output ~= self.format-failure("Non-zero wait status: $wait\n");
                         }
                     }
@@ -468,7 +467,7 @@ class Formatter::Text does Formatter {
         $output ~= "Files={ $aggregator.result-count }, Tests={ $aggregator.tests-run }$timing\n";
         my $status = $aggregator.get-status;
         $output ~= "Result: $status\n";
-        $output;
+        self.format-return($output);
     }
     method format-success(Str $output) {
         $output;
@@ -484,11 +483,9 @@ class Formatter::Text does Formatter {
         my $name = $session.header;
         if ($result.skip-all) {
             $output = self.format-return("$name skipped\n");
-        }
-        elsif ($result.has-problems($!ignore-exit)) {
+        } elsif ($result.has-problems($!ignore-exit)) {
             $output = self.format-test-failure($name, $result);
-        }
-        else {
+        } else {
             my $time = self.timer && $result.time ?? sprintf ' %8d ms', Int($result.time * 1000) !! '';
             $output = self.format-return("$name ok$time\n");
         }
@@ -507,8 +504,7 @@ class Formatter::Text does Formatter {
 
         if $result.failed == 0 {
             $output ~= self.format-failure($total ?? "All $total subtests passed " !! 'No subtests run');
-        }
-        else {
+        } else {
             $output ~= self.format-failure("Failed {$result.failed.elems}/$total subtests ");
             if (!$total) {
                 $output ~= self.format-failure("\nNo tests run!");
@@ -590,7 +586,7 @@ class Reporter::Console::Session does Session {
     }
 }
 class Reporter::Console does Reporter {
-    has Int $!lastlength = 0;
+    has Int $!last-length = 0;
     has Supplier $events = Supplier.new;
     has Reporter::Console::Session @!active;
     has Int $!tests = 0;
@@ -608,16 +604,17 @@ class Reporter::Console does Reporter {
             my @items = @!active.map(*.summary);
             my $ruler = ($header, |@items).join('  ') ~ ')===';
             $ruler = $ruler.substr(0,70) if $ruler.chars > 70;
-            $!output.print($!formatter.format-return($ruler));
+            my $output = $!formatter.format-return($ruler);
+            $!last-length = $output.chars;
+            $!output.print($output);
         }
         multi receive('update', Str $name, Str $header, Int $number, Int $plan) {
             return if self.volume <= Quiet;
             if @!active.elems == 1 {
                 my $status = ($header, $number, '/', $plan // '?').join('');
                 $!output.print($!formatter.format-return($status));
-                $!lastlength = $status.chars + 1;
-            }
-            else {
+                $!last-length = $status.chars;
+            } else {
                 output-ruler($number == 1);
             }
         }
@@ -627,7 +624,9 @@ class Reporter::Console does Reporter {
         }
         multi receive('result', Reporter::Console::Session $session, TAP::Result $result) {
             return if self.volume <= Quiet;
-            $!output.print($!formatter.format-return(' ' x $!lastlength) ~ $!formatter.format-result($session, $result));
+            my $output = $!formatter.format-result($session, $result);
+            $!output.print($!formatter.format-return(' ' x $!last-length) ~ $output);
+            $!last-length = $output.chars;
             @!active = @!active.grep(* !=== $session);
             output-ruler(True) if @!active.elems > 1;
         }
@@ -660,7 +659,7 @@ class Reporter::Console does Reporter {
 }
 
 my class State does TAP::Entry::Handler {
-    has Range $.allowed-versions = 12 .. 13;
+    has Range $.allowed-versions = 12 .. 14;
     has Int $!tests-planned;
     has Int $!tests-run = 0;
     has Int $!passed = 0;
@@ -673,6 +672,7 @@ my class State does TAP::Entry::Handler {
     has Int $!skipped = 0;
     has Int $!unknowns = 0;
     has Bool $!skip-all = False;
+    has Bool $!strict;
 
     has Promise $.bailout;
     has Int $!seen-lines = 0;
@@ -692,19 +692,16 @@ my class State does TAP::Entry::Handler {
     multi method handle-entry(TAP::Version $entry) {
         if $!seen-lines {
             self!add-error('Seen version declaration mid-stream');
-        }
-        elsif $entry.version !~~ $!allowed-versions {
+        } elsif $entry.version !~~ $!allowed-versions {
             self!add-error("Version must be in range $!allowed-versions");
-        }
-        else {
+        } else {
             $!version = $entry.version;
         }
     }
     multi method handle-entry(TAP::Plan $plan) {
         if $!seen-plan != Unseen {
             self!add-error('Seen a second plan');
-        }
-        else {
+        } else {
             $!tests-planned = $plan.tests;
             $!seen-plan = $!tests-run ?? After !! Before;
             $!skip-all = $plan.skip-all;
@@ -723,8 +720,7 @@ my class State does TAP::Entry::Handler {
         my $usable-number = $found-number // $expected-number;
         if $test.is-ok {
             $!passed++;
-        }
-        else {
+        } else {
             @!failed.push($usable-number);
         }
         ($test.ok ?? $!actual-passed !! $!actual-failed)++;
@@ -741,13 +737,20 @@ my class State does TAP::Entry::Handler {
     multi method handle-entry(TAP::Bailout $entry) {
         if $!bailout.defined {
             $!bailout.break($entry);
-        }
-        else {
+        } else {
             $!done.break($entry);
         }
     }
     multi method handle-entry(TAP::Unknown $) {
         $!unknowns++;
+        if $!strict {
+            self!add-error('Seen invalid TAP in strict mode');
+        }
+    }
+    multi method handle-entry(TAP::Pragma $entry) {
+        if $entry.identifiers<strict>:exists {
+            $!strict = $entry.identifiers<strict>;
+        }
     }
     multi method handle-entry(TAP::Entry $entry) {
     }
@@ -809,13 +812,9 @@ class Source::Supply does Source {
 
 class Async {
     has Str $.name;
-    has Run $!run handles <kill events>;
-    has State $!state;
-    has Promise $.waiter;
-
-    submethod BUILD(Str :$!name, State :$!state, Run :$!run) {
-        $!waiter = Promise.allof($!state.done, $!run.process);
-    }
+    has Run $!run handles <kill events> is built;
+    has State $!state is built;
+    has Promise $.waiter is built(False) = Promise.allof($!state.done, $!run.process);
 
     multi get_runner(Source::Proc $proc) {
         my $async = Proc::Async.new($proc.path, $proc.args);
@@ -836,7 +835,7 @@ class Async {
                 $async.bind-stderr($devnull);
             }
             when IO::Handle:D {
-                $async.stderr.lines(:close).act({ $err.say($_) });
+                $async.bind-stderr($err);
             }
             when Supplier:D {
                 $async.stderr.act({ $err.emit($_) }, :done({ $err.done }), :quit({ $err.quit($^reason) }));
@@ -890,7 +889,7 @@ class Harness {
     }
     class SourceHandler::Raku does SourceHandler::Proc {
         has Str:D $.path = $*EXECUTABLE.absolute;
-        has Str(IO) @.incdirs;
+        has @.incdirs;
         method make-source(Str:D $name, Any:D :$err, IO::Path:D :$cwd, :@include-dirs = (), *%) {
             my @dirs = map { normalize-path($_, $cwd) }, |@include-dirs, |@!incdirs;
             my @args = |@dirs.map("-I" ~ *), $name;
@@ -906,6 +905,9 @@ class Harness {
     class SourceHandler::Exec does SourceHandler::Proc {
         has Str:D $.path is required;
         has @.args;
+        method new (*@ ($path, *@args)) {
+            self.bless(:$path, :@args);
+        }
         method can-handle(Str $name) {
             1;
         }
@@ -940,13 +942,14 @@ class Harness {
     has Bool:D $.loose = $*PERL.compiler.version before 2017.09;
     has Bool $.color;
 
-    class Run {
-        has Promise $.waiter handles <result>;
-        has Promise $!killed;
-        submethod BUILD (Promise :$!waiter, Promise :$!killed) {
-        }
+    class Run does Awaitable {
+        has Promise $!waiter handles <result> is built;
+        has Promise $!bailout is built;
         method kill(Any:D $reason) {
-            $!killed.break($reason);
+            $!bailout.break($reason);
+        }
+        method get-await-handle(--> Awaitable::Handle:D) {
+            return $!waiter.promise.get-await-handle;
         }
     }
 
@@ -982,24 +985,20 @@ class Harness {
     my sub get-color(Bool $color, %env-options, Output $output) {
         with $color {
             return $color;
-        }
-        orwith %env-options<c> {
+        } orwith %env-options<c> {
             return True;
-        }
-        orwith %*ENV<HARNESS_COLOR> {
+        } orwith %*ENV<HARNESS_COLOR> {
             return ?%*ENV<HARNESS_COLOR>;
-        }
-        orwith %*ENV<NO_COLOR> {
+        } orwith %*ENV<NO_COLOR> {
             return False;
-        }
-        else {
+        } else {
             state @safe-terminals = <xterm eterm vte konsole color>;
             return $output.terminal && (%*ENV<TERM> // '') ~~ / @safe-terminals /;
         }
     }
 
     method run(*@names, IO(Str) :$cwd = $*CWD, OutVal :$out = $!output, ErrValue :$err = $!err, *%handler-args) {
-        my $killed = Promise.new;
+        my $bailout = Promise.new;
         my $aggregator = self.make-aggregator;
         my $output = make-output($out);
         my $formatter-class = get-color($!color, %!env-options, $output) ?? Formatter::Color !! Formatter::Text;
@@ -1009,23 +1008,23 @@ class Harness {
 
         my @working;
         my $waiter = start {
-            my $int = $!trap ?? sigint().tap({ $killed.break("Interrupted"); $int.close(); }) !! Tap;
+            my $int = $!trap ?? sigint().tap({ $bailout.break("Interrupted"); $int.close(); }) !! Tap;
             my $begin = now;
             try {
                 for @names -> $name {
                     my $path = normalize-path($name, $cwd);
                     my $session = $reporter.open-test($path);
                     my $source = @!handlers.max(*.can-handle($name)).make-source($path, :$err, :$cwd, |%handler-args);
-                    my $parser = TAP::Async.new(:$source, :$killed, :$!loose);
+                    my $parser = TAP::Async.new(:$source, :$bailout, :$!loose);
                     $session.listen($parser.events);
                     self.add-handlers($parser.events, $output);
                     @working.push({ :$parser, :$session, :done($parser.waiter) });
                     next if @working < $!jobs;
-                    await Promise.anyof(@working»<done>, $killed);
+                    await Promise.anyof(@working»<done>, $bailout);
                     reap-finished();
                 }
                 while @working {
-                    await Promise.anyof(@working»<done>, $killed);
+                    await Promise.anyof(@working»<done>, $bailout);
                     reap-finished();
                 }
                 CATCH {
@@ -1035,7 +1034,7 @@ class Harness {
                     }
                 }
             }
-            $reporter.summarize($aggregator, ?$killed, now - $begin) if !$killed || $!trap;
+            $reporter.summarize($aggregator, ?$bailout, now - $begin) if !$bailout || $!trap;
             $int.close if $int;
             $aggregator;
         }
@@ -1045,14 +1044,13 @@ class Harness {
                 if $done {
                     $aggregator.add-result($parser.result);
                     $session.close-test($parser.result);
-                }
-                else {
+                } else {
                     @new-working.push($current);
                 }
             }
             @working = @new-working;
         }
-        Run.new(:$waiter, :$killed);
+        Run.new(:$waiter, :$bailout);
     }
 }
 
